"""Accessor module."""
import datetime
import typing as t

import molmass
import numpy as np
import pint
import xarray as xr

from .units import to_quantity
from .units import ureg


def molecular_mass(m: str) -> pint.Quantity:  # type: ignore[type-arg]
    """Return the average molecular mass of a molecule.

    Parameters
    ----------
    m: str
        Molecule formula.

    Returns
    -------
    quantity
        Average molecular mass.
    """
    return ureg.Quantity(float(molmass.Formula(m).mass), "dalton")


@xr.register_dataset_accessor("joseki")  # type: ignore[no-untyped-call]
class JosekiAccessor:  # pragma: no cover
    """Joseki accessor."""

    def __init__(self, xarray_obj):  # type: ignore[no-untyped-def]
        self._obj = xarray_obj

    @property
    def column_number_density(
        self,
    ) -> t.Dict[str, pint.Quantity]:  # type: ignore[type-arg]
        r"""Compute column number density.

        The column number density is given by:

        .. math::
           N_{\mathrm{M}} = \int n_{\mathrm{M}}(z) \, \mathrm{d} z

        with

        .. math::

           n_{\mathrm{M}}(z) = x_{\mathrm{M}}(z) \, n(z)

        where

        * :math:`z` is the altitude,
        * :math:`x_{\mathrm{M}}(z)` is the volume mixing ratio of molecule M
          at altitude :math:`z`,
        * :math:`n(z)` is the air number density at altitude :math:`z`,
        * :math:`n_{\mathrm{M}}(z)` is the number density of molecule M at
          altitude :math:`z`.

        If the dataset has a ``z_bounds`` coordinate, the integral is computed
        using the centered rectangle method, where the ``z`` coordinate
        corresponds to the rectangle centers.

        If the dataset does not have a ``z_bounds`` coordinate, the integration
        is performed using the trapezoidal rule.

        Returns
        -------
        dict
            A mapping of molecule and column number density.
        """
        ds = self._obj
        try:
            with xr.set_options(keep_attrs=True):  # type: ignore[no-untyped-call]
                dz = to_quantity(ds.z_bounds.diff(dim="zbv", n=1).squeeze())

            n = to_quantity(ds.n)

            _column_number_density = {}
            for m in ds.m.values:
                xm = to_quantity(ds.x.sel(m=m))
                _column_number_density[m] = (
                    (xm * n * dz).sum().to_base_units()
                )  # integrate using the centered rectangle rule

            return _column_number_density

        except AttributeError:  # z_bounds attribute does not exist

            _column_number_density = {}
            for m in ds.m.values:
                integral = (ds.x.sel(m=m) * ds.n).integrate(
                    coord="z"
                )  # integrate  using the trapeziodal rule
                units = " ".join([ds[var].attrs["units"] for var in ["x", "n", "z"]])
                _column_number_density[m] = (
                    integral.values * ureg.Unit(units)
                ).to_base_units()

            return _column_number_density

    @property
    def column_mass_density(
        self,
    ) -> t.Dict[str, pint.Quantity]:  # type: ignore[type-arg]
        r"""Compute column mass density.

        The column mass density is given by:

        .. math::

           \sigma_{\mathrm{M}} = N_{\mathrm{M}} \, m_{\mathrm{M}}

        where

        * :math:`N_{\mathrm{M}}` is the column number density of molecule M,
        * :math:`m_{\mathrm{M}}` is the molecular mass of molecule M.

        Returns
        -------
        dict
            A mapping of molecule and column mass density.
        """
        _column_number_density = self.column_number_density
        return {
            m: (molecular_mass(m) * _column_number_density[m]).to("kg/m^2")
            for m in self._obj.m.values
        }

    @property
    def number_density_at_sea_level(
        self,
    ) -> t.Dict[str, pint.Quantity]:  # type: ignore[type-arg]
        """Compute number density at sea level.

        Returns
        -------
        dict:
            A mapping of molecule and number density at sea level.
        """
        ds = self._obj
        n = to_quantity(ds.n.isel(z=0))
        return {m: (to_quantity(ds.x.sel(m=m).isel(z=0)) * n) for m in ds.m.values}

    @property
    def mass_density_at_sea_level(
        self,
    ) -> t.Dict[str, pint.Quantity]:  # type: ignore[type-arg]
        """Compute mass mass density at sea level.

        Returns
        -------
        dict:
            A mapping of molecule and mass density at sea level.
        """
        _number_density_at_sea_level = self.number_density_at_sea_level
        return {
            m: (molecular_mass(m) * _number_density_at_sea_level[m]).to("kg/m^3")
            for m in self._obj.m.values
        }

    def rescale(self, factors: t.MutableMapping[str, float]) -> None:
        """Rescale molecules concentration in atmospheric profile.

        Parameters
        ----------
        factors
            A mapping of molecule and scaling factor.

        Raises
        ------
        ValueError
            When the rescaling factors are invalid, i.e. they lead to a profile
            where the volume mixing ratios sum is larger than 1.
        """
        ds = self._obj
        f = xr.DataArray(
            [factors[m] if m in factors else 1.0 for m in ds.m.values],
            coords={"m": ds.m},
            dims="m",
        )
        x_new = ds.x * f

        if np.any(x_new.sum(dim="m") > 1.0):
            raise ValueError(
                "The volume mixing ratio sum is larger than one at at least"
                " one altitude."
            )

        ds["x"].values = x_new

        now = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        for m in factors.keys():
            ds.attrs["history"] += (
                f"\n{now} - rescaled {m}'s volume mixing ratio using a scaling "
                f"factor of {round(factors[m], 3)}"
            )

        ds.attrs.update(dict(rescaled="True"))



"""Command-line interface."""
import pathlib
import typing as t

import click

from .core import Identifier
from .core import make

IDENTIFIER_CHOICES = [identifier.value for identifier in Identifier]

INTERPOLATION_METHOD_CHOICES = [
    "linear",
    "nearest",
    "nearest-up",
    "zero",
    "slinear",
    "quadratic",
    "cubic",
    "previous",
    "next",
]


@click.command()
@click.option(
    "--file-name",
    "-f",
    help="Output file name.",
    default="ds.nc",
    show_default=True,
    type=click.Path(writable=True),
)
@click.option(
    "--identifier",
    "-i",
    help="Atmospheric profile identifier.",
    required=True,
    type=click.Choice(
        choices=IDENTIFIER_CHOICES,
        case_sensitive=True,
    ),
)
@click.option(
    "--altitudes",
    "-a",
    help=(
        "Path to level altitudes data file. The data file is read with "
        ":meth:`numpy.loadtxt`."
    ),
    default=None,
    show_default=True,
)
@click.option(
    "--altitude-units",
    "-u",
    help="Altitude units",
    default="km",
    show_default=True,
)
@click.option(
    "--represent-in-cells",
    "-r",
    help=(
        "Compute the cells representation of the atmospheric profile. The "
        "initial altitude values are used to define the altitude bounds of "
        "each cell. The pressure, temperature, number density and mixing "
        "ratio fields are interpolated at the cells' center altitudes."
    ),
    is_flag=True,
)
@click.option(
    "--p-interp-method",
    "-p",
    help="Pressure interpolation method.",
    type=click.Choice(
        INTERPOLATION_METHOD_CHOICES,
        case_sensitive=True,
    ),
    default="linear",
    show_default=True,
)
@click.option(
    "--t-interp-method",
    "-t",
    help="Temperature interpolation method.",
    type=click.Choice(
        INTERPOLATION_METHOD_CHOICES,
        case_sensitive=True,
    ),
    default="linear",
    show_default=True,
)
@click.option(
    "--n-interp-method",
    "-n",
    help="Number density interpolation method.",
    type=click.Choice(
        INTERPOLATION_METHOD_CHOICES,
        case_sensitive=True,
    ),
    default="linear",
    show_default=True,
)
@click.option(
    "--x-interp-method",
    "-x",
    help="Volume mixing ratios interpolation method.",
    type=click.Choice(
        INTERPOLATION_METHOD_CHOICES,
        case_sensitive=True,
    ),
    default="linear",
    show_default=True,
)
@click.version_option()
def main(
    file_name: str,
    identifier: str,
    altitudes: t.Optional[str],
    altitude_units: str,
    represent_in_cells: bool,
    p_interp_method: str,
    t_interp_method: str,
    n_interp_method: str,
    x_interp_method: str,
) -> None:
    """Joseki command-line interface."""
    ds = make(
        identifier=Identifier(identifier),
        altitudes=altitudes if altitudes is None else pathlib.Path(altitudes),
        altitude_units=altitude_units,
        p_interp_method=p_interp_method,
        t_interp_method=t_interp_method,
        n_interp_method=n_interp_method,
        x_interp_method=x_interp_method,
        represent_in_cells=represent_in_cells,
    )
    ds.to_netcdf(file_name)


if __name__ == "__main__":
    main(prog_name="joseki")  # pragma: no cover



"""Units module."""
import pint
import xarray as xr

ureg = pint.UnitRegistry()

# define ppmv unit
ureg.define("ppmv = 1e-6 * m^3 / m^3")

# define dobson unit
ureg.define("dobson_unit = 2.687e20 * meter^-2 = du = dobson = dobson_units")


def to_quantity(da: xr.DataArray) -> pint.Quantity:  # type: ignore [type-arg]
    """Convert a :class:`~xarray.DataArray` to a :class:`~pint.Quantity`.

    The array's ``attrs`` metadata mapping must contain a ``units`` field.

    .. note:: This function can also be used on coordinate variables.

    Parameters
    ----------
    da: :class:`~xarray.DataArray`
        :class:`~xarray.DataArray` instance which will be converted.

    Returns
    -------
    :class:`~pint.Quantity`:
        The corresponding :class:`~pint.Quantity`.

    Raises
    ------
    ValueError:
        If the :class:`~xarray.DataArray`'s ``attrs`` does not contain a ``units`` key.
    """
    try:
        units = da.attrs["units"]
    except KeyError as e:
        raise ValueError("this DataArray has no 'units' metadata field") from e
    else:
        return ureg.Quantity(da.values, units)



"""Module to read AFGL 1986 data files."""
import enum
import importlib.resources as pkg_resources

import numpy as np
import pandas as pd
import xarray as xr

from .data import afgl_1986
from .units import ureg
from .util import make_data_set


class Identifier(enum.Enum):
    """AFGL 1986 atmospheric profile identifier enumeration."""

    TROPICAL = "tropical"
    MIDLATITUDE_SUMMER = "midlatitude_summer"
    MIDLATITUDE_WINTER = "midlatitude_winter"
    SUBARCTIC_SUMMER = "subarctic_summer"
    SUBARCTIC_WINTER = "subarctic_winter"
    US_STANDARD = "us_standard"


SOURCE = (
    "Atmospheric model (U.S. Standard Atmosphere) adapted from "
    "satellite data and/or dynamical-photochemical analyses."
)

REFERENCE = (
    "Anderson, G.P. and Chetwynd J.H. and Clough S.A. and "
    "Shettle E.P. and Kneizys F.X., AFGL Atmospheric "
    "Constituent Profiles (0-120km), 1986, Air Force "
    "Geophysics Laboratory, AFGL-TR-86-0110, "
    "https://ui.adsabs.harvard.edu/abs/1986afgl.rept.....A/abstract"
)

TABLE_2_DATA_FILES = (
    "table_2a.csv",
    "table_2b.csv",
    "table_2c.csv",
    "table_2d.csv",
)

DATA_FILES = {
    Identifier.TROPICAL: ("table_1a.csv", *TABLE_2_DATA_FILES),
    Identifier.MIDLATITUDE_SUMMER: ("table_1b.csv", *TABLE_2_DATA_FILES),
    Identifier.MIDLATITUDE_WINTER: ("table_1c.csv", *TABLE_2_DATA_FILES),
    Identifier.SUBARCTIC_SUMMER: ("table_1d.csv", *TABLE_2_DATA_FILES),
    Identifier.SUBARCTIC_WINTER: ("table_1e.csv", *TABLE_2_DATA_FILES),
    Identifier.US_STANDARD: ("table_1f.csv", *TABLE_2_DATA_FILES),
}


def parse(identifier: Identifier) -> pd.DataFrame:
    """Parse table data files for a given atmospheric profile.

    Read the relevant raw data files corresponding to the atmospheric profile.
    These raw data files correspond to tables 1 and 2 from the
    technical report *AFGL Atmospheric Constituent Profiles (0-120 km)*,
    Anderson et al., 1986
    :cite:`Anderson1986AtmosphericConstituentProfiles`.
    Each atmospheric profile has 5 tables, i.e. 5 raw data files, associated
    to it.
    Only the first of these tables is specific to each atmospheric profile.
    All 5 raw data files are read into :class:`~pandas.DataFrame` objects and
    then concatenated after dropping the duplicate columns.

    Parameters
    ----------
    identifier: Identifier
        Atmospheric profile identifier.

    Returns
    -------
    :class:`~pandas.DataFrame`
        Atmospheric profile data set.
    """
    files = DATA_FILES[identifier]
    dataframes = []
    for file in files:
        with pkg_resources.path(afgl_1986, file) as path:
            dataframes.append(pd.read_csv(path))
    dataframes[1] = dataframes[1].drop(["H2O", "O3", "N2O", "CO", "CH4"], axis=1)
    for i in range(1, 5):
        dataframes[i] = dataframes[i].drop("z", axis=1)

    return pd.concat(dataframes, axis=1)


def to_xarray(
    df: pd.DataFrame,
    identifier: Identifier,
    additional_molecules: bool = True,
    **kwargs: str,
) -> xr.Dataset:
    """Convert :meth:`parse`'s output to a :class:`~xarray.Dataset`.

    Use the ``z`` column of the output pandas.DataFrame of read_raw_data
    as data coordinate and all other columns as data variables.
    All data variables and coordinates of the returned xarray.Dataset are
    associated metadata (standard name, long name and units).
    Raw data units are documented in the technical report *AFGL Atmospheric
    Constituent Profiles (0-120 km)*, Anderson et al., 1986
    :cite:`Anderson1986AtmosphericConstituentProfiles`.
    Data set attributes are added.

    Parameters
    ----------
    df: :class:`~pandas.DataFrame`
        Atmospheric profile data.

    identifier: Identifier
        Atmospheric profile identifier.

    additional_molecules: bool
        If ``True``, include molecules 8-28 as numbered in
        :cite:`Anderson1986AtmosphericConstituentProfiles`.
        Else, discard molecules 8-28.

    kwargs: str
        Additional arguments passed to :meth:`util.make_data_set`.

    Returns
    -------
    :class:`~xarray.Dataset`
        Atmospheric profile data set.
    """
    # list molecules
    # molecules labels correspond to column with upper case first letter in
    # raw data DataFrames
    molecules = []
    for column in df.columns:
        if column[0].isupper():
            molecules.append(column)

    # level altitudes
    z = ureg.Quantity(df.z.values, "km")

    # air pressures
    p = ureg.Quantity(df.p.values, "millibar").to("Pa")

    # air temperatures
    t = ureg.Quantity(df.t.values, "K")

    # air number density
    n = ureg.Quantity(df.n.values, "cm^-3").to("m^-3")

    # mixing ratios
    x_values = []
    for s in molecules:
        xs = df[s].values * 1e-6  # raw data mixing ratios are in ppmv
        x_values.append(xs)
    x = ureg.Quantity(np.array(x_values), "")  # type: ignore[var-annotated]

    ds: xr.Dataset = make_data_set(  # type: ignore
        p=p,
        t=t,
        n=n,
        x=x,
        z=z,
        m=molecules,
        title=f"AFGL (1986) {identifier.value.replace('_', '-')} atmospheric profile",
        source=SOURCE,
        references=REFERENCE,
        **kwargs,
    )
    if additional_molecules:
        return ds
    else:
        return ds.isel(m=range(1, 8))


def read(identifier: Identifier, additional_molecules: bool = True) -> xr.Dataset:
    """Read data files for a given atmospheric profile.

    Chain calls to :meth:`parse` and :meth:`to_xarray`.

    Parameters
    ----------
    identifier: Identifier
        Atmospheric profile identifier.
        See :class:`.Identifier` for possible values.

    additional_molecules: bool
        If ``True``, include molecules 8-28 as numbered in
        :cite:`Anderson1986AtmosphericConstituentProfiles`.
        Else, discard molecules 8-28.

    Returns
    -------
    :class:`~xarray.Dataset`
        Atmospheric profile data set.
    """
    df = parse(identifier=identifier)
    return to_xarray(
        df=df,
        identifier=identifier,
        additional_molecules=additional_molecules,
        func_name="joseki.afgl_1986.read",
        operation="data set creation",
    )



"""Joseki."""
from . import accessor
from .core import Identifier  # pyflakes.ignore
from .core import make  # pyflakes.ignore
from .units import ureg as unit_registry


__all__ = [
    "accessor",
    "Identifier",
    "make",
    "unit_registry",
]



"""Core module."""
import enum
import pathlib
import typing as t

import numpy as np
import pint
import xarray as xr
from scipy import interpolate

from .afgl_1986 import Identifier as AFGL1986Identifier
from .afgl_1986 import read as afgl_1986_read
from .rfm import Identifier as RFMIdentifier
from .rfm import read as rfm_read
from .units import to_quantity
from .units import ureg
from .util import datetime_utcnow_stripped
from .util import make_data_set


class Identifier(enum.Enum):
    """Atmospheric profile data set identifier."""

    AFGL_1986_TROPICAL = "afgl_1986-tropical"
    AFGL_1986_MIDLATITUDE_SUMMER = "afgl_1986-midlatitude_summer"
    AFGL_1986_MIDLATITUDE_WINTER = "afgl_1986-midlatitude_winter"
    AFGL_1986_SUBARCTIC_SUMMER = "afgl_1986-subarctic_summer"
    AFGL_1986_SUBARCTIC_WINTER = "afgl_1986-subarctic_winter"
    AFGL_1986_US_STANDARD = "afgl_1986-us_standard"
    RFM_WIN = "rfm-win"
    RFM_SUM = "rfm-sum"
    RFM_DAY = "rfm-day"
    RFM_NGT = "rfm-ngt"
    RFM_EQU = "rfm-equ"
    RFM_DAY_IMK = "rfm-day_imk"
    RFM_NGT_IMK = "rfm-ngt_imk"
    RFM_SUM_IMK = "rfm-sum_imk"
    RFM_WIN_IMK = "rfm-win_imk"
    RFM_MLS = "rfm-mls"
    RFM_MLW = "rfm-mlw"
    RFM_SAS = "rfm-sas"
    RFM_SAW = "rfm-saw"
    RFM_STD = "rfm-std"
    RFM_TRO = "rfm-tro"


def convert_to_identifier(identifier: str) -> Identifier:
    """Convert identifier's value to Identifier object."""
    for i in Identifier:
        if i.value == identifier:
            return i
    raise ValueError(f"Unknown identifier '{identifier}'")


def interp(
    ds: xr.Dataset,
    z_new: pint.Quantity,  # type: ignore[type-arg]
    p_interp_method: str = "linear",
    t_interp_method: str = "linear",
    n_interp_method: str = "linear",
    x_interp_method: str = "linear",
) -> xr.Dataset:
    """Interpolate atmospheric profile on new altitudes.

    Parameters
    ----------
    ds: :class:`~xarray.Dataset`
        Atmospheric profile to interpolate.

    z_new: :class:`~pint.Quantity`
        Altitudes values at which to interpolate the atmospheric profile.

    p_interp_method: str
        Pressure interpolation method.

    t_interp_method: str
        Temperature interpolation method.

    n_interp_method: str
        Number density interpolation method.

    x_interp_method: str
        Volume mixing ratio interpolation method.

    Returns
    -------
    :class:`~xarray.Dataset`
        Interpolated atmospheric profile.
    """
    z_units = ds.z.attrs["units"]
    z_new_values = z_new.m_as(z_units)

    # Interpolate pressure
    fp = interpolate.interp1d(
        ds.z.values, ds.p.values, kind=p_interp_method, bounds_error=True
    )
    p_new = ureg.Quantity(fp(z_new_values), ds.p.attrs["units"])

    # Interpolate temperature
    ft = interpolate.interp1d(
        ds.z.values, ds.t.values, kind=t_interp_method, bounds_error=True
    )
    t_new = ureg.Quantity(ft(z_new_values), ds.t.attrs["units"])

    # Interpolate number density
    fn = interpolate.interp1d(
        ds.z.values, ds.n.values, kind=n_interp_method, bounds_error=True
    )
    n_new = ureg.Quantity(fn(z_new_values), ds.n.attrs["units"])

    # Interpolate volume mixing ratio
    x_new = (
        ds.x.interp(
            z=z_new_values,
            method=x_interp_method,
            kwargs=dict(bounds_error=True),
        ).values
        * ureg.dimensionless
    )

    # Reform data set
    interpolated = make_data_set(  # type: ignore
        p=p_new,
        t=t_new,
        n=n_new,
        x=x_new,
        z=z_new,
        m=list(ds.m.values),
        func_name="joseki.core.interp",
        operation="data set interpolation",
        **ds.attrs,
    )
    return interpolated  # type: ignore


def represent_profile_in_cells(
    ds: xr.Dataset,
    p_interp_method: str = "linear",
    t_interp_method: str = "linear",
    n_interp_method: str = "linear",
    x_interp_method: str = "linear",
) -> xr.Dataset:
    """Compute the cells representation of the atmospheric profile.

    Atmosphere cells (or layers) are defined by two consecutive altitude values.
    The layer's center altitude is defined as the average of these two values.
    The pressure, temperature, number density and mixing ratio fields are
    interpolated at these layer' center altitude values.
    In the new atmospheric profile, the 'zn' coordinate is removed, a layer'
    center altitude coordinate 'z' is added and a data variable 'z_bounds'
    indicating the altitude bounds of each layer, is added.


    Parameters
    ----------
    ds: :class:`~xarray.Dataset`
        Initial atmospheric profile.

    p_interp_method: str
        Pressure interpolation method.

    t_interp_method: str
        Temperature interpolation method.

    n_interp_method: str
        Number density interpolation method.

    x_interp_method: str
        Volume mixing ratio interpolation method.

    Returns
    -------
    :class:`~xarray.Dataset`
        Cells representation of the atmospheric profile.
    """
    # if the profile is already represented in cells, do nothing
    if ds.z.standard_name == "layer_center_altitude":
        return ds

    z_nodes = to_quantity(ds.z)
    z_centers = (z_nodes[:-1] + z_nodes[1:]) / 2.0
    interpolated: xr.Dataset = interp(  # type: ignore
        ds=ds,
        z_new=z_centers,
        p_interp_method=p_interp_method,
        t_interp_method=t_interp_method,
        n_interp_method=n_interp_method,
        x_interp_method=x_interp_method,
    )
    interpolated.z.attrs = dict(
        standard_name="layer_center_altitude",
        long_name="layer center altitude",
        units="km",
    )
    z_bounds = np.array(
        [z_nodes.magnitude[:-1], z_nodes.magnitude[1:]]
    )  # type: ignore[var-annotated]
    interpolated = interpolated.assign(
        z_bounds=(
            ("zbv", "z"),
            z_bounds,
            dict(
                standard_name="altitude",
                long_name="altitude",
                units="km",
            ),
        )
    )
    interpolated.attrs.update(
        history=interpolated.history + f"\n{datetime_utcnow_stripped()} "
        "- data set coords update - joseki.core.represent_profile_in_cells"
    )

    return interpolated


def make(
    identifier: t.Union[str, Identifier],
    altitudes: t.Optional[pathlib.Path] = None,
    altitude_units: str = "km",
    represent_in_cells: bool = False,
    p_interp_method: str = "linear",
    t_interp_method: str = "linear",
    n_interp_method: str = "linear",
    x_interp_method: str = "linear",
    **kwargs: t.Any,
) -> xr.Dataset:
    """Make atmospheric profile data set.

    This method creates an atmospheric profile data set corresponding to the
    specified identifier.
    It allows to interpolate the atmospheric profile on a different
    level altitude mesh as well as to set the layer center altitude as the data
    set's main coordinate.

    Parameters
    ----------
    identifier: Identifier
        Atmospheric profile identifier.

    altitudes: pathlib.Path
        Altitudes data file.
        If ``None``, the original atmospheric profile altitudes are used.

    altitude_units: str
        Altitude units.

    represent_in_cells: bool
        If ``True``, compute the cells representation of the atmospheric profile.

    p_interp_method: str
        Pressure interpolation method.

    t_interp_method: str
        Temperature interpolation method.

    n_interp_method: str
        Number density interpolation method.

    x_interp_method: str
        Volume mixing ratios interpolation method

    kwargs: Any
        Parameters passed to lower-level readers.

    Returns
    -------
    :class:`~xarray.Dataset`
        Atmospheric profile.
    """
    if isinstance(identifier, str):
        identifier = convert_to_identifier(identifier=identifier)
    group_name, identifier_name = identifier.value.split("-")
    if group_name == "afgl_1986":
        ds = afgl_1986_read(identifier=AFGL1986Identifier(identifier_name), **kwargs)
    if group_name == "rfm":
        ds = rfm_read(identifier=RFMIdentifier(identifier_name), **kwargs)

    if altitudes is not None:
        z_new_values = np.loadtxt(  # type: ignore[no-untyped-call]
            fname=altitudes,
            dtype=float,
            comments="#",
        )
        ds = interp(  # type: ignore
            ds=ds,
            z_new=ureg.Quantity(z_new_values, altitude_units),
            p_interp_method=p_interp_method,
            t_interp_method=t_interp_method,
            n_interp_method=n_interp_method,
            x_interp_method=x_interp_method,
        )

    if represent_in_cells:
        ds = represent_profile_in_cells(
            ds=ds,
            p_interp_method=p_interp_method,
            t_interp_method=t_interp_method,
            n_interp_method=n_interp_method,
            x_interp_method=x_interp_method,
        )

    return ds



"""Module to read atmospheric profiles distributed with Reference Forward Model.

Reference Forward Model (RFM) website: http://eodg.atm.ox.ac.uk/RFM/.
"""
import enum
import importlib.resources as pkg_resources
import typing as t
from datetime import datetime

import numpy as np
import pint
import requests
import xarray as xr
from scipy.constants import physical_constants

from .data import rfm
from .units import ureg
from .util import make_data_set
from .util import to_chemical_formula

SOURCE = "unknown"

REFERENCE = "unknown"

DESCRIPTION = {
    "win": "MIPAS (2001) polar winter",
    "sum": "MIPAS (2001) polar summer",
    "day": "MIPAS (2001) mid-latitude daytime",
    "ngt": "MIPAS (2001) mid-latitude nighttime",
    "equ": "MIPAS (2001) equatorial",
    "day_200km": "Extended MIPAS Model Atmospheres (2011) - Mid-Latitude, day-time",
    "ngt_200km": "Extended MIPAS Model Atmospheres (2011) - Mid-Latitude, night-time",
    "win_200km": "Extended MIPAS Model Atmospheres (2011) - Polar Winter",
    "sum_200km": "Extended MIPAS Model Atmospheres (2011) - Polar Summer",
    "equ_200km": "Extended MIPAS Model Atmospheres (2011) - Equatorial, day-time",
    "mls": "AFGL (1986) Mid-latitude summer",
    "mlw": "AFGL (1986) Mid-latitude winter",
    "sas": "AFGL (1986) Sub-arctic summer",
    "saw": "AFGL (1986) Sub-arctic winter",
    "std": "AFGL (1986) U.S. Standard",
    "tro": "AFGL (1986) Tropical",
}


class Identifier(enum.Enum):
    """RFM atmospheric profile identifier enumeration."""

    WIN = "win"
    SUM = "sum"
    DAY = "day"
    NGT = "ngt"
    EQU = "equ"
    DAY_200KM = "day_200km"
    NGT_200KM = "ngt_200km"
    WIN_200KM = "win_200km"
    SUM_200KM = "sum_200km"
    EQU_200KM = "equ_200km"
    MLS = "mls"
    MLW = "mlw"
    SAS = "sas"
    SAW = "saw"
    STD = "std"
    TRO = "tro"


# Boltzmann constant
K = ureg.Quantity(*physical_constants["Boltzmann constant"][:2])


def _parse_units(s: str) -> str:
    """Parse units."""
    if s.startswith("[") and s.endswith("]"):
        units = s[1:-1]
        if units == "mb":
            return "millibar"
        else:
            return units
    else:
        raise ValueError(f"Cannot parse units '{s}'")


def _parse_var_name(n: str) -> str:
    """Parse variable name."""
    translate = {"HGT": "z", "PRE": "p", "TEM": "t"}
    if n in translate.keys():
        return translate[n]
    else:
        return to_chemical_formula(n)


def _parse_var_line(s: str) -> t.Tuple[str, str]:
    """Parse a line with the declaration of a variable and its units."""
    parts = s[1:].strip().split()
    if len(parts) == 2:
        var_name, units_s = parts
    elif len(parts) == 3:
        var_name, _, units_s = parts
    else:
        raise ValueError(f"Invalid line format: {s}")
    var = _parse_var_name(var_name)
    units = _parse_units(units_s)
    return var, units


def _parse_values_line(s: str) -> t.List[str]:
    """Parse a line with numeric values."""
    if "," in s:  # delimiter is comma and whitespace combined
        s_strip = s.strip()
        if s_strip[-1] == ",":
            s_strip = s_strip[:-1]
        return [x.strip() for x in s_strip.split(",")]
    else:  # delimiter is whitespace
        return s.split()


def _parse_content(lines: t.List[str]) -> t.Dict[str, pint.Quantity]:  # type: ignore
    """Parse lines."""
    iterator = iter(lines)
    line = next(iterator)

    quantities: t.Dict[str, pint.Quantity] = {}  # type: ignore [type-arg]

    def _add_to_quantities(quantity: pint.Quantity, name: str) -> None:  # type: ignore
        if quantity.units == "ppmv":
            quantities[name] = quantity.to("dimensionless")
        else:
            quantities[name] = quantity

    var: str = ""
    units: str = ""
    values: t.List[str] = []
    while line != "*END":
        if line.startswith("!"):
            pass  # this is a comment, ignore the line
        elif line.startswith("*"):
            # convert previously read values (if any) and units to quantity
            if len(values) > 0:
                quantity = ureg.Quantity(
                    np.array(values, dtype=float),
                    units,
                )  # type: ignore[var-annotated]
                _add_to_quantities(quantity=quantity, name=var)

            # this is a variable line, parse variable name and units
            var, units = _parse_var_line(line)

            # following lines are the variables values so prepare a variable
            # to store the values
            values = []
        else:
            if "!" in line:
                # this the line with the number of profile levels, ignore it
                pass
            else:
                # this line contains variable values
                values += _parse_values_line(line)
        line = next(iterator)

    # include last array of values before the '*END' line
    quantity = ureg.Quantity(np.array(values, dtype=float), units)
    _add_to_quantities(quantity=quantity, name=var)

    return quantities


def read_file_content(identifier: Identifier) -> t.Tuple[str, t.Dict[str, str]]:
    """
    Read atmospheric profile data file content.

    Parameters
    ----------
    identifier: Identifier
        Atmospheric profile identifier.
        See :class:`.Identifier` for possible values.

    Returns
    -------
    tuple:
        file content, URL, URL date.
    """
    file_name = identifier.value.split("-")[-1]
    return _read_file_content(file_name=file_name)


def _read_file_content(file_name: str) -> t.Tuple[str, t.Dict[str, str]]:
    """
    Read data file content.

    Parameters
    ----------
    file_name: str
        Atmospheric data file name.

    Returns
    -------
    tuple:
        file content, URL, URL date.
    """
    try:
        url = f"http://eodg.atm.ox.ac.uk/RFM/atm/{file_name}.atm"
        url_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        response = requests.get(url)
        content = response.text
        return content, dict(url=url, url_date=url_date)
    except requests.ConnectionError:
        file = f"{file_name}.atm"
        with pkg_resources.path(rfm, file) as path:
            with open(path, "r") as f:
                content = f.read()
        return content, dict()


def read_additional_molecules(
    identifier: Identifier,
) -> t.Tuple[t.Dict[str, pint.Quantity], t.Dict[str, str]]:  # type: ignore [type-arg]
    """
    Read additional molecules data file.

    Parameters
    ----------
    identifier: Identifier
        Atmospheric profile identifier.
        See :class:`.Identifier` for possible values.

    Raises
    ------
    ValueError
        If the profile does not have additional molecules.

    Returns
    -------
    tuple of dict of str and :class:`~pint.Quantity` and dict of str and str:
        Additional molecules parsed content, URL information.
    """
    if identifier in [
        Identifier.DAY,
        Identifier.EQU,
        Identifier.NGT,
        Identifier.SUM,
        Identifier.WIN,
    ]:
        add_molecules_name = "extra"
    elif identifier in [
        Identifier.MLS,
        Identifier.MLW,
        Identifier.SAS,
        Identifier.SAW,
        Identifier.STD,
        Identifier.TRO,
    ]:
        add_molecules_name = "minor"
    else:
        raise ValueError

    content, url_info = _read_file_content(file_name=add_molecules_name)
    parsed_content = _parse_content(content.splitlines())
    return parsed_content, url_info


def read(
    identifier: Identifier, additional_molecules: t.Optional[bool] = False
) -> xr.Dataset:
    """Read RFM atmospheric data files.

    Try to read the data from http://eodg.atm.ox.ac.uk/RFM/atm/
    If that fails, reads archived data files in ``src/joseki/data/rfm/``.
    The archived raw data files were downloaded from
    http://eodg.atm.ox.ac.uk/RFM/atm/ on June 6th, 2021.

    Parameters
    ----------
    identifier: Identifier
        Atmospheric profile identifier.
        See :class:`.Identifier` for possible values.

    additional_molecules: bool
        Set to ``True`` to include the additional molecules to the atmospheric
        profile.

    Returns
    -------
    :class:`~xarray.Dataset`
        Atmospheric profile.
    """
    content, url_info = read_file_content(identifier=identifier)
    url = url_info.get("url")
    url_date = url_info.get("url_date")
    quantities = _parse_content(content.splitlines())

    z = quantities.pop("z")
    p = quantities.pop("p")
    t = quantities.pop("t")
    n = p / (K * t)  # perfect gas equation
    molecules = list(quantities.keys())
    x = (
        np.array([quantities[molecule].magnitude for molecule in molecules])
        * ureg.dimensionless
    )  # type: ignore[var-annotated]

    if additional_molecules:
        extra_quantities, extra_url_info = read_additional_molecules(
            identifier=identifier
        )
        extra_z = extra_quantities.pop("z")
        extra_m = list(extra_quantities.keys())
        extra_x = np.array(
            [extra_quantities[s].magnitude for s in extra_m]
        )  # type: ignore[var-annotated]

        # initial molecules
        da = xr.DataArray(
            x.m,
            dims=["molecules", "z"],
            coords={"molecules": molecules, "z": z.magnitude},
        )

        # additional molecules
        da_extra = xr.DataArray(
            extra_x,
            dims=["molecules", "z"],
            coords={
                "molecules": extra_m,
                "z": extra_z.m_as(z.units),
            },
        )

        # interpolate additional molecules mixing ratio on altitude mesh used
        # for initial molecules mixing ratio:
        da_extra_interp = np.exp(
            np.log(da_extra).interp(z=z.m_as(z.units))  # type: ignore
        )

        # concatenate initial and additional molecules
        da_total = xr.concat([da, da_extra_interp], dim="molecules")
        molecules = list(da_total.molecules.values)
        x = da_total.values * ureg.dimensionless

    ds = make_data_set(
        p=p,
        t=t,
        n=n,
        x=x,
        z=z,
        m=molecules,
        func_name="joseki.rfm.read",
        operation="data set creation",
        title=f"RFM {DESCRIPTION[identifier.value]} atmospheric profile",
        source=SOURCE,
        references=REFERENCE,
        url=url,
        url_date=url_date,
    )
    return ds  # type: ignore



"""Utility module."""
import datetime
import typing as t

import pint
import xarray as xr

from .units import ureg


def datetime_utcnow_stripped() -> datetime.datetime:
    """Return current UTC with seconds and microseconds stripped."""
    return datetime.datetime.utcnow().replace(second=0, microsecond=0)


def make_data_set(
    p: pint.Quantity,  # type: ignore[type-arg]
    t: pint.Quantity,  # type: ignore[type-arg]
    n: pint.Quantity,  # type: ignore[type-arg]
    x: pint.Quantity,  # type: ignore[type-arg]
    z: pint.Quantity,  # type: ignore[type-arg]
    m: t.List[str],
    convention: str = "CF-1.8",
    title: str = "unknown",
    history: t.Optional[str] = None,
    func_name: str = "unknown",
    operation: str = "unknown",
    source: str = "unknown",
    references: str = "unknown",
    url: t.Optional[str] = None,
    url_date: t.Optional[str] = None,
) -> xr.Dataset:
    """Make an atmospheric profile data set.

    Parameters
    ----------
    p: :class:`~pint.Quantity`
        Pressure values.

    t: :class:`~pint.Quantity`
        Temperature values.

    n: :class:`~pint.Quantity`
        Number density values.

    x: :class:`~pint.Quantity`
        Volume mixing ratios values.

    z: :class:`~pint.Quantity`
        Altitude values.

    m: list of str
        Molecules.

    convention: str
        Metadata convention.

    title: str
        A succinct description of what is in the dataset.

    history: str, optional
        Provides an audit trail for modifications to the original data.

    func_name: str
        Name of the calling function.

    operation: str
        Name of the operation performed on the data set.

    source: str
        The method of production of the original data.

    references: str
        Published or web-based references that describe the data or methods
        used to produce it.

    url: str, optional
        Data URL.

    url_date: str, optional
        Date and time of last access to URL.

    Returns
    -------
    :class:`~xarray.Dataset`
        Atmospheric profile.
    """
    if history is None:
        history = f"{datetime_utcnow_stripped()} - {operation} - {func_name}"
    else:
        history += f"\n{datetime_utcnow_stripped()} - {operation} - {func_name}"
    attrs = dict(
        convention=convention,
        title=title,
        history=history,
        source=source,
        references=references,
    )
    if url is not None:
        attrs.update(dict(url=url))
    if url_date is not None:
        attrs.update(dict(url_date=url_date))

    return xr.Dataset(
        data_vars=dict(
            p=(
                "z",
                p.m_as(ureg.pascal),
                dict(
                    standard_name="air_pressure",
                    long_name="air pressure",
                    units="Pa",
                ),
            ),
            t=(
                "z",
                t.m_as(ureg.kelvin),
                dict(
                    standard_name="air_temperature",
                    long_name="air temperature",
                    units="K",
                ),
            ),
            n=(
                "z",
                n.m_as(ureg.m**-3),
                dict(
                    standard_name="air_number_density",
                    long_name="air number density",
                    units="m^-3",
                ),
            ),
            x=(
                ("m", "z"),
                x.m_as(ureg.dimensionless),
                dict(
                    standard_name="volume_mixing_ratio",
                    long_name="volume mixing ratio",
                    units="",
                ),
            ),
        ),
        coords=dict(
            z=(
                "z",
                z.m_as(ureg.kilometer),
                dict(
                    standard_name="altitude",
                    long_name="altitude",
                    units="km",
                ),
            ),
            m=(
                "m",
                m,
                dict(
                    standard_name="molecule",
                    long_name="molecule",
                ),
            ),
        ),
        attrs=attrs,  # type: ignore
    )


CFC_FORMULAE = {
    "CCl3F": ("Freon-11", "F11", "R-11", "CFC-11"),
    "CCl2F2": ("Freon-12", "F12", "R-12", "CFC-12"),
    "CClF3": ("Freon-13", "F13", "R-13", "CFC-13"),
    "CF4": ("Freon-14", "F14", "CFC-14"),
    "CHClF2": ("Freon-22", "F22", "HCFC-22"),
    "CHCl2F": ("Freon-21", "F21", "HCFC-21"),
    "C2Cl3F3": ("Freon-113", "F113", "CFC-113"),
    "C2Cl2F4": ("Freon-114", "F114", "CFC-114"),
}


def to_chemical_formula(name: str) -> str:
    """Convert to chemical formula.

    If molecule name is unknown, returns name unchanged.

    Parameters
    ----------
    name: str
        Molecule name.

    Returns
    -------
    str:
        Molecule formula.
    """
    try:
        return translate_cfc(name)
    except ValueError:
        return name


def translate_cfc(name: str) -> str:
    """Convert chlorofulorocarbon name to corresponding chemical formula.

    Parameters
    ----------
    name: str
        Chlorofulorocarbon name.

    Returns
    -------
    str:
        Chlorofulorocarbon chemical formula.

    Raises
    ------
    ValueError:
        If the name does not match a known chlorofulorocarbon.
    """
    for formula, names in CFC_FORMULAE.items():
        if name in names:
            return formula
    raise ValueError("Unknown chlorofulorocarbon {name}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['joseki', 'joseki.data', 'joseki.data.afgl_1986', 'joseki.data.rfm']

package_data = \
{'': ['*']}

install_requires = \
['Pint>=0.17',
 'click>=7.0',
 'molmass>=2021.6.18',
 'netCDF4>=1.5.7',
 'numpy>=1.22.1',
 'pandas>=1.2.4',
 'requests>=2.25.1',
 'scipy>=1.6.3',
 'xarray>=0.18.2']

entry_points = \
{'console_scripts': ['joseki = joseki.__main__:main']}

setup_kwargs = {
    'name': 'joseki',
    'version': '1.3.0',
    'description': 'Joseki',
    'long_description': "Joseki\n======\n\n|PyPI| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/joseki.svg\n   :target: https://pypi.org/project/joseki/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/joseki\n   :target: https://pypi.org/project/joseki\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/joseki\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/joseki/latest.svg?label=Read%20the%20Docs\n   :target: https://joseki.readthedocs.io/\n   :alt: Read the documentation at https://joseki.readthedocs.io/\n.. |Tests| image:: https://github.com/nollety/joseki/workflows/Tests/badge.svg\n   :target: https://github.com/nollety/joseki/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/nollety/joseki/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/nollety/joseki\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nFeatures\n--------\n\n* *AFGL Atmospheric Constituent Profiles (0-120 km)* data sets in\n  `csv <https://en.wikipedia.org/wiki/Comma-separated_values>`_ and\n  `netCDF <https://www.unidata.ucar.edu/software/netcdf/>`_ formats.\n* Atmospheric profiles from `RFM <http://eodg.atm.ox.ac.uk/RFM/>`_\n* Atmospheric profile interpolation on altitude.\n* Projection of the nodes-based profile on the corresponding centers-based altitude mesh.\n* Command-line interface.\n* Python API.\n\n\nRequirements\n------------\n\n* Python 3.8+\n\n\nInstallation\n------------\n\nYou can install *Joseki* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install joseki\n\n\nPython API\n----------\n\nPlease see the `User Guide`_ for basic use.\nFor more details, refer to the `API Reference`_.\n\n\nCommand-line interface\n----------------------\n\nPlease see the `Command-line Reference`_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the MIT_ license,\n*Joseki* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_'s\n`Hypermodern Python Cookiecutter`_ template.\n\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT: http://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/nollety/joseki/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Command-line Reference: https://joseki.readthedocs.io/en/latest/cli.html\n.. _User Guide: https://joseki.readthedocs.io/en/latest/user_guide.html\n.. _API Reference: https://joseki.readthedocs.io/en/latest/reference.html\n",
    'author': 'Yvan Nollet',
    'author_email': 'yvan.nollet@rayference.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nollety/joseki',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
